@@ -24,10 +24,15 @@
 # SOFTWARE.
 
 # AUTHORS
-# Rachid RIAD & Hadrien TITEUX
+# Rachid RIAD, Hadrien TITEUX, Léopold FAVRE
 
-from .continuum import Continuum, GammaResults
+from .continuum import Continuum, GammaResults, Unit
 from .alignment import Alignment, UnitaryAlignment
 from .dissimilarity import (PositionalDissimilarity,
                             CategoricalDissimilarity,
                             CombinedCategoricalDissimilarity)
+from .sampler import (AbstractContinuumSampler,
+                      ShuffleContinuumSampler,
+                      StatisticalContinuumSampler)
+from .notebook import show_continuum, show_alignment
+from .cst import CorpusShufflingTool
